from ._core import (
    DomainModel,
    DomainModelCore,
    DomainModelApply,
    DomainModelList,
    DomainRelationApply,
    ResourcesApply,
    ResourcesApplyResult,
)
from ._alert import Alert, AlertApply, AlertApplyList, AlertFields, AlertList, AlertTextFields
from ._basic_bid_matrix import (
    BasicBidMatrix,
    BasicBidMatrixApply,
    BasicBidMatrixApplyList,
    BasicBidMatrixFields,
    BasicBidMatrixList,
    BasicBidMatrixTextFields,
)
from ._bid_document import (
    BidDocument,
    BidDocumentApply,
    BidDocumentApplyList,
    BidDocumentFields,
    BidDocumentList,
    BidDocumentTextFields,
)
from ._bid_matrix import (
    BidMatrix,
    BidMatrixApply,
    BidMatrixApplyList,
    BidMatrixFields,
    BidMatrixList,
    BidMatrixTextFields,
)
from ._bid_method import (
    BidMethod,
    BidMethodApply,
    BidMethodApplyList,
    BidMethodFields,
    BidMethodList,
    BidMethodTextFields,
)
from ._multi_scenario_matrix import (
    MultiScenarioMatrix,
    MultiScenarioMatrixApply,
    MultiScenarioMatrixApplyList,
    MultiScenarioMatrixFields,
    MultiScenarioMatrixList,
    MultiScenarioMatrixTextFields,
)
from ._price_area import (
    PriceArea,
    PriceAreaApply,
    PriceAreaApplyList,
    PriceAreaFields,
    PriceAreaList,
    PriceAreaTextFields,
)
from ._shop_multi_scenario import (
    SHOPMultiScenario,
    SHOPMultiScenarioApply,
    SHOPMultiScenarioApplyList,
    SHOPMultiScenarioFields,
    SHOPMultiScenarioList,
    SHOPMultiScenarioTextFields,
)
from ._water_value_based import (
    WaterValueBased,
    WaterValueBasedApply,
    WaterValueBasedApplyList,
    WaterValueBasedFields,
    WaterValueBasedList,
    WaterValueBasedTextFields,
)

BasicBidMatrix.model_rebuild()
BasicBidMatrixApply.model_rebuild()
BidDocument.model_rebuild()
BidDocumentApply.model_rebuild()
BidMatrix.model_rebuild()
BidMatrixApply.model_rebuild()
MultiScenarioMatrix.model_rebuild()
MultiScenarioMatrixApply.model_rebuild()
PriceArea.model_rebuild()
PriceAreaApply.model_rebuild()

__all__ = [
    "ResourcesApply",
    "DomainModel",
    "DomainModelCore",
    "DomainModelApply",
    "DomainModelList",
    "DomainRelationApply",
    "ResourcesApplyResult",
    "Alert",
    "AlertApply",
    "AlertList",
    "AlertApplyList",
    "AlertFields",
    "AlertTextFields",
    "BasicBidMatrix",
    "BasicBidMatrixApply",
    "BasicBidMatrixList",
    "BasicBidMatrixApplyList",
    "BasicBidMatrixFields",
    "BasicBidMatrixTextFields",
    "BidDocument",
    "BidDocumentApply",
    "BidDocumentList",
    "BidDocumentApplyList",
    "BidDocumentFields",
    "BidDocumentTextFields",
    "BidMatrix",
    "BidMatrixApply",
    "BidMatrixList",
    "BidMatrixApplyList",
    "BidMatrixFields",
    "BidMatrixTextFields",
    "BidMethod",
    "BidMethodApply",
    "BidMethodList",
    "BidMethodApplyList",
    "BidMethodFields",
    "BidMethodTextFields",
    "MultiScenarioMatrix",
    "MultiScenarioMatrixApply",
    "MultiScenarioMatrixList",
    "MultiScenarioMatrixApplyList",
    "MultiScenarioMatrixFields",
    "MultiScenarioMatrixTextFields",
    "PriceArea",
    "PriceAreaApply",
    "PriceAreaList",
    "PriceAreaApplyList",
    "PriceAreaFields",
    "PriceAreaTextFields",
    "SHOPMultiScenario",
    "SHOPMultiScenarioApply",
    "SHOPMultiScenarioList",
    "SHOPMultiScenarioApplyList",
    "SHOPMultiScenarioFields",
    "SHOPMultiScenarioTextFields",
    "WaterValueBased",
    "WaterValueBasedApply",
    "WaterValueBasedList",
    "WaterValueBasedApplyList",
    "WaterValueBasedFields",
    "WaterValueBasedTextFields",
]



from __future__ import annotations

import itertools
import logging
import re
from collections import defaultdict
from dataclasses import asdict, dataclass
from graphlib import TopologicalSorter
from pathlib import Path
from typing import Any, Generic, TypeVar

import yaml
from cognite.client import CogniteClient
from cognite.client.data_classes import filters
from cognite.client.data_classes.data_modeling import (
    ContainerApplyList,
    DataModelApply,
    DataModelApplyList,
    MappedPropertyApply,
    NodeApplyList,
    NodeId,
    SpaceApplyList,
    ViewApplyList,
)
from cognite.client.data_classes.data_modeling.views import SingleHopConnectionDefinitionApply, ViewApply

logger = logging.getLogger(__name__)

_DMS_DIR = Path(__file__).parent / "dms"


@dataclass
class Schema:
    _shared_space = "power-ops-shared"
    containers: ContainerApplyList
    _views: ViewApplyList
    _data_models: DataModelApplyList
    spaces: SpaceApplyList
    node_types: NodeApplyList

    @property
    def container_views(self) -> ViewApplyList:
        # The data views are views that are a 1-1 mapping to a container
        # They are used to see what data is available in a container
        return ViewApplyList(
            [
                ViewApply(
                    space=self._shared_space,
                    external_id=f"{container.external_id}ContainerData_{container.space.replace('-','_')}",
                    version="1",
                    name=f"{container.name}ContainerData",
                    description=f"Data view for {container.name} in {container.space}",
                    properties={
                        prop: MappedPropertyApply(container=container.as_id(), container_property_identifier=prop)
                        for prop in container.properties
                    },
                )
                for container in self.containers
            ]
        )

    @property
    def container_model(self) -> DataModelApply:
        return DataModelApply(
            space=self._shared_space,
            external_id="PowerOpsContainerModel",
            version="1",
            name="PowerOpsContainers",
            description="All the data available in PowerOps",
            views=self.container_views.as_ids(),
        )

    @property
    def views(self) -> ViewApplyList:
        return self._views + self.container_views

    @property
    def data_models(self) -> DataModelApplyList:
        return self._data_models + DataModelApplyList([self.container_model])


@dataclass
class Difference:
    create: list[Any]
    changed: list[Any]
    unchanged: list[Any]
    delete: list[Any]

    def as_results(self, is_init: bool) -> list[dict]:
        if is_init:
            return (
                [{"action": "created", "model": item.as_id()} for item in self.create]
                + [{"action": "changed", "model": item.as_id()} for item in self.changed]
                + [{"action": "skipped-delete", "model": item.as_id()} for item in self.delete]
            )
        else:
            return [
                {"action": "deleted", "model": item.as_id()}
                for item in itertools.chain(self.delete, self.changed, self.unchanged)
            ]


T_ResourceList = TypeVar("T_ResourceList")


class DataModelAPI(Generic[T_ResourceList]):
    def __init__(self, class_api: Any, client: CogniteClient):
        self._client = client
        self._class_api = class_api

    @property
    def name(self) -> str:
        return type(self._class_api).__module__.rsplit(".", maxsplit=1)[-1]

    def apply(self, items: T_ResourceList) -> Any:
        return self._class_api.apply(items)

    def retrieve(self, spaces: SpaceApplyList) -> T_ResourceList:
        retrieved: T_ResourceList | None = None
        for space in spaces:
            if retrieved is None:
                retrieved = self._class_api.list(space=space.space)
            else:
                retrieved.extend(self._class_api.list(space=space.space))
        retrieved_apply = retrieved.as_apply() if retrieved else []
        if isinstance(retrieved_apply, ViewApplyList):
            # Need to remove properties that originate from other views
            view_by_id = {view.as_id(): view for view in retrieved_apply}
            for view in retrieved_apply:
                prop_to_pop: list[str] = []
                views_to_check = list(view.implements or [])
                while views_to_check:
                    view_to_check = views_to_check.pop()
                    prop_to_pop.extend(view_by_id[view_to_check].properties)
                    views_to_check.extend(view_by_id[view_to_check].implements or [])
                for prop in prop_to_pop:
                    view.properties.pop(prop)
        return retrieved_apply

    def differences(self, cdf: T_ResourceList, local: T_ResourceList) -> Difference:
        cdf_resources_by_id = {item.as_id(): item for item in cdf}
        local_resources_by_id = {item.as_id(): item for item in local}
        created = [item for item in local if item.as_id() not in cdf_resources_by_id]
        deleted = [item for item in cdf if item.as_id() not in local_resources_by_id]
        changed = []
        unchanged = []
        for item in local:
            if item.as_id() in cdf_resources_by_id:
                if cdf_resources_by_id[item.as_id()].dump() == item.dump():
                    unchanged.append(item)
                else:
                    changed.append(item)
        return Difference(create=created, changed=changed, unchanged=unchanged, delete=deleted)

    def delete(self, ids: list[Any]) -> None:
        return self._class_api.delete(ids)


class SpaceAPI(DataModelAPI[SpaceApplyList]):
    def retrieve(self, spaces: SpaceApplyList) -> SpaceApplyList:
        all_spaces = self._client.data_modeling.spaces.list()
        return SpaceApplyList([space.as_apply() for space in all_spaces if space.space.startswith("power-ops")])

    def delete(self, ids: list[Any]) -> None:
        for space in ids:
            if space == "power-ops-types":
                return  # Never delete the types space
            data_models = self._client.data_modeling.data_models.list(space=space)
            self._client.data_modeling.data_models.delete(data_models.as_ids())
            views = self._client.data_modeling.views.list(space=space)
            self._client.data_modeling.views.delete(views.as_ids())
            containers = self._client.data_modeling.containers.list(space=space)
            self._client.data_modeling.containers.delete(containers.as_ids())
            is_space = filters.Equals(["edge", "space"], space)
            edges = self._client.data_modeling.instances.list("edge", filter=is_space)
            self._client.data_modeling.instances.delete(edges.as_ids())
            is_space = filters.Equals(["node", "space"], space)
            nodes = self._client.data_modeling.instances.list(filter=is_space)
            self._client.data_modeling.instances.delete(nodes.as_ids())
            if nodes or edges or containers or views or data_models:
                print(
                    f"Deleted {space}: {len(nodes)} nodes, {len(edges)} edges, {len(containers)} containers, "
                    f"{len(views)} views, {len(data_models)} data models"
                )

        self._client.data_modeling.spaces.delete(ids)


class DataModelLoader:
    def __init__(self):
        self._source_dir = _DMS_DIR
        self._config_file = _DMS_DIR / "config.yaml"
        if not self._config_file.exists():
            raise ValueError(f"Missing config file {self._config_file!s}. Expected to be in {_DMS_DIR!s}")
        self._config = yaml.safe_load(self._config_file.read_text())

    def load(self) -> Schema:
        resources_by_type = defaultdict(list)
        for filepath in self._source_dir.glob("**/*.yaml"):
            if match := re.match(r".*(?P<type>(space|view|container|node|data_model))\.yaml$", filepath.name):
                loaded = self._load_file(filepath)
                if isinstance(loaded, list):
                    resources_by_type[match.group("type")].extend(loaded)
                else:
                    resources_by_type[match.group("type")].append(loaded)

        return Schema(
            containers=ContainerApplyList.load(resources_by_type["container"]),
            _views=ViewApplyList.load(resources_by_type["view"]),
            _data_models=DataModelApplyList.load(resources_by_type["data_model"]),
            spaces=SpaceApplyList.load(resources_by_type["space"]),
            node_types=NodeApplyList.load(resources_by_type["node"]),
        )

    def _load_file(self, filepath: Path) -> dict[str, Any] | list[dict[str, Any]]:
        file_contents = filepath.read_text()
        for variable, value in self._config.items():
            file_contents = file_contents.replace(f"{{{{{ variable }}}}}", value)
        if "{{" in file_contents:
            errors = []
            for line_no, line in enumerate(file_contents.split("\n"), start=1):
                if "{{" in line:
                    position = line.index("{{")
                    errors.append(f"Line {line_no} - col {position}: {line[position:position+20]!r}")
            errors = "\n".join(errors) if errors else ""
            raise ValueError(f"Unresolved variables in {filepath.relative_to(Path.cwd())} near: {errors}")
        return yaml.safe_load(file_contents)

    @classmethod
    def validate(cls, schema: Schema):
        defined_spaces = {space.space for space in schema.spaces}
        defined_containers = {container.as_id() for container in schema.containers}
        defined_views = {view.as_id() for view in schema._views}
        defined_node_types = {node_type.as_id() for node_type in schema.node_types}
        defined_interfaces = {parent for view in schema._views for parent in view.implements or []}
        properties_by_container = {container.as_id(): set(container.properties) for container in schema.containers}

        referred_spaces = defaultdict(list)
        referred_views = defaultdict(list)
        referred_containers = defaultdict(list)
        referred_node_types = defaultdict(list)
        view_missing_filters = []
        non_existent_container_properties = []
        for container in schema.containers:
            referred_spaces[container.space].append(container.as_id())
        for view in schema._views:
            ref_view_id = view.as_id()
            referred_spaces[view.space].append(ref_view_id)
            if isinstance(view.filter, filters.Equals):
                dumped = view.filter.dump()["equals"]["value"]
                try:
                    if "space" in dumped and "externalId" in dumped:
                        node_id = NodeId(dumped["space"], dumped["externalId"])
                        referred_node_types[node_id].append(ref_view_id)
                except Exception as exc:
                    raise ValueError(
                        f"Failed to parse filter for view {view.space}.{view.external_id}.{view.version}"
                    ) from exc
            elif isinstance(view.filter, filters.In):
                dumped = view.filter.dump()["in"]["values"]
                try:
                    for value in dumped:
                        if "space" in value and "externalId" in value:
                            node_id = NodeId(value["space"], value["externalId"])
                            referred_node_types[node_id].append(ref_view_id)
                except Exception as exc:
                    raise ValueError(
                        f"Failed to parse filter for view {view.space}.{view.external_id}.{view.version}"
                    ) from exc

            if ref_view_id in defined_interfaces and not isinstance(view.filter, filters.In):
                view_missing_filters.append(ref_view_id)
            elif ref_view_id not in defined_interfaces and isinstance(view.filter, filters.Equals):
                view_missing_filters.append(ref_view_id)

            for view_id in view.implements or []:
                referred_views[view_id].append(ref_view_id)
            for prop_name, prop in view.properties.items():
                if isinstance(prop, MappedPropertyApply):
                    referred_containers[prop.container].append(ref_view_id.as_property_ref(prop_name))
                    if prop.source:
                        referred_views[prop.source].append(ref_view_id.as_property_ref(prop_name))
                    if prop.container_property_identifier not in properties_by_container.get(prop.container, set()):
                        non_existent_container_properties.append(ref_view_id.as_property_ref(prop_name))
                elif isinstance(prop, SingleHopConnectionDefinitionApply):
                    referred_node_types[NodeId(prop.type.space, prop.type.external_id)].append(
                        ref_view_id.as_property_ref(prop_name)
                    )
                    referred_views[prop.source].append(ref_view_id.as_property_ref(prop_name))
                    if prop.edge_source:
                        referred_views[prop.edge_source].append(ref_view_id.as_property_ref(prop_name))
        for node in schema.node_types:
            referred_spaces[node.space].append(node.as_id())
        for data_model in schema._data_models:
            referred_spaces[data_model.space].append(data_model.as_id())
            for view in data_model.views:
                referred_views[view].append(data_model.as_id())

        if undefined_spaces := set(referred_spaces).difference(defined_spaces):
            referred_to_by = list(itertools.chain(*(referred_spaces[space] for space in undefined_spaces)))
            raise ValueError(f"Undefined spaces: {undefined_spaces}: referred to by {referred_to_by}")
        if undefined_containers := set(referred_containers).difference(defined_containers):
            referred_to_by = list(
                itertools.chain(*(referred_containers[container] for container in undefined_containers))
            )
            raise ValueError(f"Undefined containers: {undefined_containers}: referred to by {referred_to_by}")
        if undefined_views := set(referred_views).difference(defined_views):
            referred_to_by = list(itertools.chain(*(referred_views[view] for view in undefined_views)))
            raise ValueError(f"Undefined views: {undefined_views}: referred to by {referred_to_by}")
        if undefined_node_types := set(referred_node_types).difference(defined_node_types):
            referred_to_by = list(
                itertools.chain(*(referred_node_types[node_type] for node_type in undefined_node_types))
            )
            raise ValueError(f"Undefined node types: {undefined_node_types}: referred to by {referred_to_by}")
        if non_existent_container_properties:
            message = "\n".join(map(str, non_existent_container_properties))
            raise ValueError(f"These properties in views refers to container properties that does not exist: {message}")

    @classmethod
    def deploy(cls, client: CogniteClient, schema: Schema, is_dev: bool = False) -> list[dict]:
        result = []
        apis = cls._create_apis(client)
        resources = {
            "spaces": schema.spaces,
            "containers": schema.containers,
            "views": schema.views,
            "data_models": schema.data_models,
        }
        for api in TopologicalSorter(apis).static_order():
            items = resources[api.name]
            existing = api.retrieve(schema.spaces)
            diffs = api.differences(existing, items)
            if diffs.create or diffs.changed:
                if is_dev and api.name in {"views", "data_models"} and (diffs.changed or diffs.delete):
                    print(f"Deleting {api.name}: {len(diffs.changed + diffs.delete)} changed to avoid conflicts")
                    api.delete([item.as_id() for item in diffs.changed + diffs.delete])
                api.apply(diffs.create + diffs.changed)

            result.extend(diffs.as_results(is_init=True))
            print(
                f"Deployed {api.name}: {len(diffs.create)} created, {len(diffs.changed)} changed, "
                f"{len(diffs.unchanged)} unchanged, ({len(diffs.delete)} deleted - skipped)"
            )
        return result

    @classmethod
    def destroy(cls, client: CogniteClient, schema: Schema, dry_run: bool = False) -> list[dict]:
        result = []
        apis = cls._create_apis(client)
        resources = asdict(schema)
        for api in reversed(list(TopologicalSorter(apis).static_order())):
            items = resources[api.name]
            existing = api.retrieve(schema.spaces)
            diffs = api.differences(existing, items)
            if not dry_run and (diffs.changed or diffs.unchanged or diffs.delete):
                api.delete(items.as_ids())
            total = len(diffs.changed + diffs.unchanged + diffs.delete)
            if dry_run:
                print(f"Would delete {api.name}: {total} deleted")
            else:
                print(f"Deleted {api.name}: {total} deleted")
            result.extend(diffs.as_results(is_init=False))

        return result

    @classmethod
    def _create_apis(cls, client: CogniteClient) -> dict[DataModelAPI, set[DataModelAPI]]:
        space = SpaceAPI(client.data_modeling.spaces, client)
        container = DataModelAPI[ContainerApplyList](client.data_modeling.containers, client)
        view = DataModelAPI[ViewApplyList](client.data_modeling.views, client)
        data_model = DataModelAPI[DataModelApplyList](client.data_modeling.data_models, client)
        return {space: set(), container: {space}, view: {space, container}, data_model: {space, container, view}}


if __name__ == "__main__":
    # This is here to make it easy to check that the models are valid (all variables are resolved)
    loader = DataModelLoader()
    schema = loader.load()
    print("Schema loaded")
    DataModelLoader.validate(schema)
    print("Schema validated")



from __future__ import annotations

from typing import Literal, Optional, Union

from cognite.client import data_modeling as dm

from ._core import (
    DEFAULT_INSTANCE_SPACE,
    DomainModel,
    DomainModelCore,
    DomainModelApply,
    DomainModelApplyList,
    DomainModelList,
    DomainRelationApply,
    ResourcesApply,
)


__all__ = [
    "BidMethod",
    "BidMethodApply",
    "BidMethodList",
    "BidMethodApplyList",
    "BidMethodFields",
    "BidMethodTextFields",
]


BidMethodTextFields = Literal["name"]
BidMethodFields = Literal["name"]

_BIDMETHOD_PROPERTIES_BY_FIELD = {
    "name": "name",
}


class BidMethod(DomainModel):
    """This represents the reading version of bid method.

    It is used to when data is retrieved from CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the bid method.
        name: Name for the BidMethod
        created_time: The created time of the bid method node.
        last_updated_time: The last updated time of the bid method node.
        deleted_time: If present, the deleted time of the bid method node.
        version: The version of the bid method node.
    """

    space: str = DEFAULT_INSTANCE_SPACE
    node_type: Union[dm.DirectRelationReference, None] = None
    name: Optional[str] = None

    def as_apply(self) -> BidMethodApply:
        """Convert this read version of bid method to the writing version."""
        return BidMethodApply(
            space=self.space,
            external_id=self.external_id,
            name=self.name,
        )


class BidMethodApply(DomainModelApply):
    """This represents the writing version of bid method.

    It is used to when data is sent to CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the bid method.
        name: Name for the BidMethod
        existing_version: Fail the ingestion request if the bid method version is greater than or equal to this value.
            If no existingVersion is specified, the ingestion will always overwrite any existing data for the edge (for the specified container or instance).
            If existingVersion is set to 0, the upsert will behave as an insert, so it will fail the bulk if the item already exists.
            If skipOnVersionConflict is set on the ingestion request, then the item will be skipped instead of failing the ingestion request.
    """

    space: str = DEFAULT_INSTANCE_SPACE
    node_type: Union[dm.DirectRelationReference, None] = None
    name: str

    def _to_instances_apply(
        self,
        cache: set[tuple[str, str]],
        view_by_read_class: dict[type[DomainModelCore], dm.ViewId] | None,
    ) -> ResourcesApply:
        resources = ResourcesApply()
        if self.as_tuple_id() in cache:
            return resources

        write_view = (view_by_read_class or {}).get(BidMethod, dm.ViewId("power-ops-shared", "BidMethod", "1"))

        properties = {}

        if self.name is not None:
            properties["name"] = self.name

        if properties:
            this_node = dm.NodeApply(
                space=self.space,
                external_id=self.external_id,
                existing_version=self.existing_version,
                type=self.node_type,
                sources=[
                    dm.NodeOrEdgeData(
                        source=write_view,
                        properties=properties,
                    )
                ],
            )
            resources.nodes.append(this_node)
            cache.add(self.as_tuple_id())

        return resources


class BidMethodList(DomainModelList[BidMethod]):
    """List of bid methods in the read version."""

    _INSTANCE = BidMethod

    def as_apply(self) -> BidMethodApplyList:
        """Convert these read versions of bid method to the writing versions."""
        return BidMethodApplyList([node.as_apply() for node in self.data])


class BidMethodApplyList(DomainModelApplyList[BidMethodApply]):
    """List of bid methods in the writing version."""

    _INSTANCE = BidMethodApply


def _create_bid_method_filter(
    view_id: dm.ViewId,
    name: str | list[str] | None = None,
    name_prefix: str | None = None,
    external_id_prefix: str | None = None,
    space: str | list[str] | None = None,
    filter: dm.Filter | None = None,
) -> dm.Filter | None:
    filters = []
    if name is not None and isinstance(name, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("name"), value=name))
    if name and isinstance(name, list):
        filters.append(dm.filters.In(view_id.as_property_ref("name"), values=name))
    if name_prefix:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("name"), value=name_prefix))
    if external_id_prefix:
        filters.append(dm.filters.Prefix(["node", "externalId"], value=external_id_prefix))
    if space is not None and isinstance(space, str):
        filters.append(dm.filters.Equals(["node", "space"], value=space))
    if space and isinstance(space, list):
        filters.append(dm.filters.In(["node", "space"], values=space))
    if filter:
        filters.append(filter)
    return dm.filters.And(*filters) if filters else None



from __future__ import annotations

from collections.abc import Sequence
from typing import overload

from cognite.client import CogniteClient
from cognite.client import data_modeling as dm
from cognite.client.data_classes.data_modeling.instances import InstanceAggregationResultList

from cognite.powerops.client._generated.assets.data_classes._core import DEFAULT_INSTANCE_SPACE
from cognite.powerops.client._generated.assets.data_classes import (
    DomainModelCore,
    DomainModelApply,
    ResourcesApplyResult,
    Watercourse,
    WatercourseApply,
    WatercourseFields,
    WatercourseList,
    WatercourseApplyList,
    WatercourseTextFields,
)
from cognite.powerops.client._generated.assets.data_classes._watercourse import (
    _WATERCOURSE_PROPERTIES_BY_FIELD,
    _create_watercourse_filter,
)
from ._core import (
    DEFAULT_LIMIT_READ,
    DEFAULT_QUERY_LIMIT,
    Aggregations,
    NodeAPI,
    SequenceNotStr,
    QueryStep,
    QueryBuilder,
)
from .watercourse_plants import WatercoursePlantsAPI
from .watercourse_production_obligation import WatercourseProductionObligationAPI
from .watercourse_query import WatercourseQueryAPI


class WatercourseAPI(NodeAPI[Watercourse, WatercourseApply, WatercourseList]):
    def __init__(self, client: CogniteClient, view_by_read_class: dict[type[DomainModelCore], dm.ViewId]):
        view_id = view_by_read_class[Watercourse]
        super().__init__(
            client=client,
            sources=view_id,
            class_type=Watercourse,
            class_list=WatercourseList,
            class_apply_list=WatercourseApplyList,
            view_by_read_class=view_by_read_class,
        )
        self._view_id = view_id
        self.plants_edge = WatercoursePlantsAPI(client)
        self.production_obligation = WatercourseProductionObligationAPI(client, view_id)

    def __call__(
        self,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        display_name: str | list[str] | None = None,
        display_name_prefix: str | None = None,
        min_penalty_limit: float | None = None,
        max_penalty_limit: float | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int = DEFAULT_QUERY_LIMIT,
        filter: dm.Filter | None = None,
    ) -> WatercourseQueryAPI[WatercourseList]:
        """Query starting at watercourses.

        Args:
            name: The name to filter on.
            name_prefix: The prefix of the name to filter on.
            display_name: The display name to filter on.
            display_name_prefix: The prefix of the display name to filter on.
            min_penalty_limit: The minimum value of the penalty limit to filter on.
            max_penalty_limit: The maximum value of the penalty limit to filter on.
            external_id_prefix: The prefix of the external ID to filter on.
            space: The space to filter on.
            limit: Maximum number of watercourses to return. Defaults to 25. Set to -1, float("inf") or None to return all items.
            filter: (Advanced) If the filtering available in the above is not sufficient, you can write your own filtering which will be ANDed with the filter above.

        Returns:
            A query API for watercourses.

        """
        has_data = dm.filters.HasData(views=[self._view_id])
        filter_ = _create_watercourse_filter(
            self._view_id,
            name,
            name_prefix,
            display_name,
            display_name_prefix,
            min_penalty_limit,
            max_penalty_limit,
            external_id_prefix,
            space,
            (filter and dm.filters.And(filter, has_data)) or has_data,
        )
        builder = QueryBuilder(WatercourseList)
        return WatercourseQueryAPI(self._client, builder, self._view_by_read_class, filter_, limit)

    def apply(
        self, watercourse: WatercourseApply | Sequence[WatercourseApply], replace: bool = False
    ) -> ResourcesApplyResult:
        """Add or update (upsert) watercourses.

        Note: This method iterates through all nodes and timeseries linked to watercourse and creates them including the edges
        between the nodes. For example, if any of `plants` are set, then these
        nodes as well as any nodes linked to them, and all the edges linking these nodes will be created.

        Args:
            watercourse: Watercourse or sequence of watercourses to upsert.
            replace (bool): How do we behave when a property value exists? Do we replace all matching and existing values with the supplied values (true)?
                Or should we merge in new values for properties together with the existing values (false)? Note: This setting applies for all nodes or edges specified in the ingestion call.
        Returns:
            Created instance(s), i.e., nodes, edges, and time series.

        Examples:

            Create a new watercourse:

                >>> from cognite.powerops.client._generated.assets import PowerAssetAPI
                >>> from cognite.powerops.client._generated.assets.data_classes import WatercourseApply
                >>> client = PowerAssetAPI()
                >>> watercourse = WatercourseApply(external_id="my_watercourse", ...)
                >>> result = client.watercourse.apply(watercourse)

        """
        return self._apply(watercourse, replace)

    def delete(
        self, external_id: str | SequenceNotStr[str], space: str = DEFAULT_INSTANCE_SPACE
    ) -> dm.InstancesDeleteResult:
        """Delete one or more watercourse.

        Args:
            external_id: External id of the watercourse to delete.
            space: The space where all the watercourse are located.

        Returns:
            The instance(s), i.e., nodes and edges which has been deleted. Empty list if nothing was deleted.

        Examples:

            Delete watercourse by id:

                >>> from cognite.powerops.client._generated.assets import PowerAssetAPI
                >>> client = PowerAssetAPI()
                >>> client.watercourse.delete("my_watercourse")
        """
        return self._delete(external_id, space)

    @overload
    def retrieve(self, external_id: str, space: str = DEFAULT_INSTANCE_SPACE) -> Watercourse | None:
        ...

    @overload
    def retrieve(self, external_id: SequenceNotStr[str], space: str = DEFAULT_INSTANCE_SPACE) -> WatercourseList:
        ...

    def retrieve(
        self, external_id: str | SequenceNotStr[str], space: str = DEFAULT_INSTANCE_SPACE
    ) -> Watercourse | WatercourseList | None:
        """Retrieve one or more watercourses by id(s).

        Args:
            external_id: External id or list of external ids of the watercourses.
            space: The space where all the watercourses are located.

        Returns:
            The requested watercourses.

        Examples:

            Retrieve watercourse by id:

                >>> from cognite.powerops.client._generated.assets import PowerAssetAPI
                >>> client = PowerAssetAPI()
                >>> watercourse = client.watercourse.retrieve("my_watercourse")

        """
        return self._retrieve(
            external_id,
            space,
            retrieve_edges=True,
            edge_api_name_type_direction_quad=[
                (
                    self.plants_edge,
                    "plants",
                    dm.DirectRelationReference("power-ops-types", "isSubAssetOf"),
                    "outwards",
                ),
            ],
        )

    def search(
        self,
        query: str,
        properties: WatercourseTextFields | Sequence[WatercourseTextFields] | None = None,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        display_name: str | list[str] | None = None,
        display_name_prefix: str | None = None,
        min_penalty_limit: float | None = None,
        max_penalty_limit: float | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> WatercourseList:
        """Search watercourses

        Args:
            query: The search query,
            properties: The property to search, if nothing is passed all text fields will be searched.
            name: The name to filter on.
            name_prefix: The prefix of the name to filter on.
            display_name: The display name to filter on.
            display_name_prefix: The prefix of the display name to filter on.
            min_penalty_limit: The minimum value of the penalty limit to filter on.
            max_penalty_limit: The maximum value of the penalty limit to filter on.
            external_id_prefix: The prefix of the external ID to filter on.
            space: The space to filter on.
            limit: Maximum number of watercourses to return. Defaults to 25. Set to -1, float("inf") or None to return all items.
            filter: (Advanced) If the filtering available in the above is not sufficient, you can write your own filtering which will be ANDed with the filter above.

        Returns:
            Search results watercourses matching the query.

        Examples:

           Search for 'my_watercourse' in all text properties:

                >>> from cognite.powerops.client._generated.assets import PowerAssetAPI
                >>> client = PowerAssetAPI()
                >>> watercourses = client.watercourse.search('my_watercourse')

        """
        filter_ = _create_watercourse_filter(
            self._view_id,
            name,
            name_prefix,
            display_name,
            display_name_prefix,
            min_penalty_limit,
            max_penalty_limit,
            external_id_prefix,
            space,
            filter,
        )
        return self._search(self._view_id, query, _WATERCOURSE_PROPERTIES_BY_FIELD, properties, filter_, limit)

    @overload
    def aggregate(
        self,
        aggregations: Aggregations
        | dm.aggregations.MetricAggregation
        | Sequence[Aggregations]
        | Sequence[dm.aggregations.MetricAggregation],
        property: WatercourseFields | Sequence[WatercourseFields] | None = None,
        group_by: None = None,
        query: str | None = None,
        search_properties: WatercourseTextFields | Sequence[WatercourseTextFields] | None = None,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        display_name: str | list[str] | None = None,
        display_name_prefix: str | None = None,
        min_penalty_limit: float | None = None,
        max_penalty_limit: float | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> list[dm.aggregations.AggregatedNumberedValue]:
        ...

    @overload
    def aggregate(
        self,
        aggregations: Aggregations
        | dm.aggregations.MetricAggregation
        | Sequence[Aggregations]
        | Sequence[dm.aggregations.MetricAggregation],
        property: WatercourseFields | Sequence[WatercourseFields] | None = None,
        group_by: WatercourseFields | Sequence[WatercourseFields] = None,
        query: str | None = None,
        search_properties: WatercourseTextFields | Sequence[WatercourseTextFields] | None = None,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        display_name: str | list[str] | None = None,
        display_name_prefix: str | None = None,
        min_penalty_limit: float | None = None,
        max_penalty_limit: float | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> InstanceAggregationResultList:
        ...

    def aggregate(
        self,
        aggregate: Aggregations
        | dm.aggregations.MetricAggregation
        | Sequence[Aggregations]
        | Sequence[dm.aggregations.MetricAggregation],
        property: WatercourseFields | Sequence[WatercourseFields] | None = None,
        group_by: WatercourseFields | Sequence[WatercourseFields] | None = None,
        query: str | None = None,
        search_property: WatercourseTextFields | Sequence[WatercourseTextFields] | None = None,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        display_name: str | list[str] | None = None,
        display_name_prefix: str | None = None,
        min_penalty_limit: float | None = None,
        max_penalty_limit: float | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> list[dm.aggregations.AggregatedNumberedValue] | InstanceAggregationResultList:
        """Aggregate data across watercourses

        Args:
            aggregate: The aggregation to perform.
            property: The property to perform aggregation on.
            group_by: The property to group by when doing the aggregation.
            query: The query to search for in the text field.
            search_property: The text field to search in.
            name: The name to filter on.
            name_prefix: The prefix of the name to filter on.
            display_name: The display name to filter on.
            display_name_prefix: The prefix of the display name to filter on.
            min_penalty_limit: The minimum value of the penalty limit to filter on.
            max_penalty_limit: The maximum value of the penalty limit to filter on.
            external_id_prefix: The prefix of the external ID to filter on.
            space: The space to filter on.
            limit: Maximum number of watercourses to return. Defaults to 25. Set to -1, float("inf") or None to return all items.
            filter: (Advanced) If the filtering available in the above is not sufficient, you can write your own filtering which will be ANDed with the filter above.

        Returns:
            Aggregation results.

        Examples:

            Count watercourses in space `my_space`:

                >>> from cognite.powerops.client._generated.assets import PowerAssetAPI
                >>> client = PowerAssetAPI()
                >>> result = client.watercourse.aggregate("count", space="my_space")

        """

        filter_ = _create_watercourse_filter(
            self._view_id,
            name,
            name_prefix,
            display_name,
            display_name_prefix,
            min_penalty_limit,
            max_penalty_limit,
            external_id_prefix,
            space,
            filter,
        )
        return self._aggregate(
            self._view_id,
            aggregate,
            _WATERCOURSE_PROPERTIES_BY_FIELD,
            property,
            group_by,
            query,
            search_property,
            limit,
            filter_,
        )

    def histogram(
        self,
        property: WatercourseFields,
        interval: float,
        query: str | None = None,
        search_property: WatercourseTextFields | Sequence[WatercourseTextFields] | None = None,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        display_name: str | list[str] | None = None,
        display_name_prefix: str | None = None,
        min_penalty_limit: float | None = None,
        max_penalty_limit: float | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> dm.aggregations.HistogramValue:
        """Produces histograms for watercourses

        Args:
            property: The property to use as the value in the histogram.
            interval: The interval to use for the histogram bins.
            query: The query to search for in the text field.
            search_property: The text field to search in.
            name: The name to filter on.
            name_prefix: The prefix of the name to filter on.
            display_name: The display name to filter on.
            display_name_prefix: The prefix of the display name to filter on.
            min_penalty_limit: The minimum value of the penalty limit to filter on.
            max_penalty_limit: The maximum value of the penalty limit to filter on.
            external_id_prefix: The prefix of the external ID to filter on.
            space: The space to filter on.
            limit: Maximum number of watercourses to return. Defaults to 25. Set to -1, float("inf") or None to return all items.
            filter: (Advanced) If the filtering available in the above is not sufficient, you can write your own filtering which will be ANDed with the filter above.

        Returns:
            Bucketed histogram results.

        """
        filter_ = _create_watercourse_filter(
            self._view_id,
            name,
            name_prefix,
            display_name,
            display_name_prefix,
            min_penalty_limit,
            max_penalty_limit,
            external_id_prefix,
            space,
            filter,
        )
        return self._histogram(
            self._view_id,
            property,
            interval,
            _WATERCOURSE_PROPERTIES_BY_FIELD,
            query,
            search_property,
            limit,
            filter_,
        )

    def list(
        self,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        display_name: str | list[str] | None = None,
        display_name_prefix: str | None = None,
        min_penalty_limit: float | None = None,
        max_penalty_limit: float | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
        retrieve_edges: bool = True,
    ) -> WatercourseList:
        """List/filter watercourses

        Args:
            name: The name to filter on.
            name_prefix: The prefix of the name to filter on.
            display_name: The display name to filter on.
            display_name_prefix: The prefix of the display name to filter on.
            min_penalty_limit: The minimum value of the penalty limit to filter on.
            max_penalty_limit: The maximum value of the penalty limit to filter on.
            external_id_prefix: The prefix of the external ID to filter on.
            space: The space to filter on.
            limit: Maximum number of watercourses to return. Defaults to 25. Set to -1, float("inf") or None to return all items.
            filter: (Advanced) If the filtering available in the above is not sufficient, you can write your own filtering which will be ANDed with the filter above.
            retrieve_edges: Whether to retrieve `plants` external ids for the watercourses. Defaults to True.

        Returns:
            List of requested watercourses

        Examples:

            List watercourses and limit to 5:

                >>> from cognite.powerops.client._generated.assets import PowerAssetAPI
                >>> client = PowerAssetAPI()
                >>> watercourses = client.watercourse.list(limit=5)

        """
        filter_ = _create_watercourse_filter(
            self._view_id,
            name,
            name_prefix,
            display_name,
            display_name_prefix,
            min_penalty_limit,
            max_penalty_limit,
            external_id_prefix,
            space,
            filter,
        )

        return self._list(
            limit=limit,
            filter=filter_,
            retrieve_edges=retrieve_edges,
            edge_api_name_type_direction_quad=[
                (
                    self.plants_edge,
                    "plants",
                    dm.DirectRelationReference("power-ops-types", "isSubAssetOf"),
                    "outwards",
                ),
            ],
        )



from __future__ import annotations

import datetime
from typing import TYPE_CHECKING, Literal, Optional, Union

from cognite.client import data_modeling as dm
from pydantic import Field

from ._core import (
    DEFAULT_INSTANCE_SPACE,
    DomainModel,
    DomainModelCore,
    DomainModelApply,
    DomainModelApplyList,
    DomainModelList,
    DomainRelationApply,
    ResourcesApply,
)

if TYPE_CHECKING:
    from ._alert import Alert, AlertApply
    from ._bid_matrix import BidMatrix, BidMatrixApply
    from ._bid_method import BidMethod, BidMethodApply
    from ._price_area import PriceArea, PriceAreaApply


__all__ = [
    "BidDocument",
    "BidDocumentApply",
    "BidDocumentList",
    "BidDocumentApplyList",
    "BidDocumentFields",
    "BidDocumentTextFields",
]


BidDocumentTextFields = Literal["name"]
BidDocumentFields = Literal["name", "delivery_date", "start_calculation", "end_calculation", "is_complete"]

_BIDDOCUMENT_PROPERTIES_BY_FIELD = {
    "name": "name",
    "delivery_date": "deliveryDate",
    "start_calculation": "startCalculation",
    "end_calculation": "endCalculation",
    "is_complete": "isComplete",
}


class BidDocument(DomainModel):
    """This represents the reading version of bid document.

    It is used to when data is retrieved from CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the bid document.
        name: Unique name for a given instance of a Bid Document. A combination of name, priceArea, date and startCalculation.
        delivery_date: The date of the Bid.
        start_calculation: Timestamp of when the Bid calculation workflow started.
        end_calculation: Timestamp of when the Bid calculation workflow completed.
        is_complete: Indicates that the Bid calculation workflow has completed (although has not necessarily succeeded).
        alerts: An array of calculation level Alerts.
        price_area: The price area field.
        method: The method field.
        total: The total field.
        partials: The partial field.
        created_time: The created time of the bid document node.
        last_updated_time: The last updated time of the bid document node.
        deleted_time: If present, the deleted time of the bid document node.
        version: The version of the bid document node.
    """

    space: str = DEFAULT_INSTANCE_SPACE
    node_type: Union[dm.DirectRelationReference, None] = dm.DirectRelationReference(
        "power-ops-types", "DayAheadBidDocument"
    )
    name: Optional[str] = None
    delivery_date: Optional[datetime.date] = Field(None, alias="deliveryDate")
    start_calculation: Optional[datetime.datetime] = Field(None, alias="startCalculation")
    end_calculation: Optional[datetime.datetime] = Field(None, alias="endCalculation")
    is_complete: Optional[bool] = Field(None, alias="isComplete")
    alerts: Union[list[Alert], list[str], None] = Field(default=None, repr=False)
    price_area: Union[PriceArea, str, dm.NodeId, None] = Field(None, repr=False, alias="priceArea")
    method: Union[BidMethod, str, dm.NodeId, None] = Field(None, repr=False)
    total: Union[BidMatrix, str, dm.NodeId, None] = Field(None, repr=False)
    partials: Union[list[BidMatrix], list[str], None] = Field(default=None, repr=False)

    def as_apply(self) -> BidDocumentApply:
        """Convert this read version of bid document to the writing version."""
        return BidDocumentApply(
            space=self.space,
            external_id=self.external_id,
            name=self.name,
            delivery_date=self.delivery_date,
            start_calculation=self.start_calculation,
            end_calculation=self.end_calculation,
            is_complete=self.is_complete,
            alerts=[alert.as_apply() if isinstance(alert, DomainModel) else alert for alert in self.alerts or []],
            price_area=self.price_area.as_apply() if isinstance(self.price_area, DomainModel) else self.price_area,
            method=self.method.as_apply() if isinstance(self.method, DomainModel) else self.method,
            total=self.total.as_apply() if isinstance(self.total, DomainModel) else self.total,
            partials=[
                partial.as_apply() if isinstance(partial, DomainModel) else partial for partial in self.partials or []
            ],
        )


class BidDocumentApply(DomainModelApply):
    """This represents the writing version of bid document.

    It is used to when data is sent to CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the bid document.
        name: Unique name for a given instance of a Bid Document. A combination of name, priceArea, date and startCalculation.
        delivery_date: The date of the Bid.
        start_calculation: Timestamp of when the Bid calculation workflow started.
        end_calculation: Timestamp of when the Bid calculation workflow completed.
        is_complete: Indicates that the Bid calculation workflow has completed (although has not necessarily succeeded).
        alerts: An array of calculation level Alerts.
        price_area: The price area field.
        method: The method field.
        total: The total field.
        partials: The partial field.
        existing_version: Fail the ingestion request if the bid document version is greater than or equal to this value.
            If no existingVersion is specified, the ingestion will always overwrite any existing data for the edge (for the specified container or instance).
            If existingVersion is set to 0, the upsert will behave as an insert, so it will fail the bulk if the item already exists.
            If skipOnVersionConflict is set on the ingestion request, then the item will be skipped instead of failing the ingestion request.
    """

    space: str = DEFAULT_INSTANCE_SPACE
    node_type: Union[dm.DirectRelationReference, None] = dm.DirectRelationReference(
        "power-ops-types", "DayAheadBidDocument"
    )
    name: Optional[str] = None
    delivery_date: datetime.date = Field(alias="deliveryDate")
    start_calculation: Optional[datetime.datetime] = Field(None, alias="startCalculation")
    end_calculation: Optional[datetime.datetime] = Field(None, alias="endCalculation")
    is_complete: Optional[bool] = Field(None, alias="isComplete")
    alerts: Union[list[AlertApply], list[str], None] = Field(default=None, repr=False)
    price_area: Union[PriceAreaApply, str, dm.NodeId, None] = Field(None, repr=False, alias="priceArea")
    method: Union[BidMethodApply, str, dm.NodeId, None] = Field(None, repr=False)
    total: Union[BidMatrixApply, str, dm.NodeId, None] = Field(None, repr=False)
    partials: Union[list[BidMatrixApply], list[str], None] = Field(default=None, repr=False)

    def _to_instances_apply(
        self,
        cache: set[tuple[str, str]],
        view_by_read_class: dict[type[DomainModelCore], dm.ViewId] | None,
    ) -> ResourcesApply:
        resources = ResourcesApply()
        if self.as_tuple_id() in cache:
            return resources

        write_view = (view_by_read_class or {}).get(
            BidDocument, dm.ViewId("power-ops-day-ahead-bid", "BidDocument", "1")
        )

        properties = {}

        if self.name is not None:
            properties["name"] = self.name

        if self.delivery_date is not None:
            properties["deliveryDate"] = self.delivery_date.isoformat()

        if self.start_calculation is not None:
            properties["startCalculation"] = self.start_calculation.isoformat(timespec="milliseconds")

        if self.end_calculation is not None:
            properties["endCalculation"] = self.end_calculation.isoformat(timespec="milliseconds")

        if self.is_complete is not None:
            properties["isComplete"] = self.is_complete

        if self.price_area is not None:
            properties["priceArea"] = {
                "space": self.space if isinstance(self.price_area, str) else self.price_area.space,
                "externalId": self.price_area if isinstance(self.price_area, str) else self.price_area.external_id,
            }

        if self.method is not None:
            properties["method"] = {
                "space": self.space if isinstance(self.method, str) else self.method.space,
                "externalId": self.method if isinstance(self.method, str) else self.method.external_id,
            }

        if self.total is not None:
            properties["total"] = {
                "space": self.space if isinstance(self.total, str) else self.total.space,
                "externalId": self.total if isinstance(self.total, str) else self.total.external_id,
            }

        if properties:
            this_node = dm.NodeApply(
                space=self.space,
                external_id=self.external_id,
                existing_version=self.existing_version,
                type=self.node_type,
                sources=[
                    dm.NodeOrEdgeData(
                        source=write_view,
                        properties=properties,
                    )
                ],
            )
            resources.nodes.append(this_node)
            cache.add(self.as_tuple_id())

        edge_type = dm.DirectRelationReference("power-ops-types", "calculationIssue")
        for alert in self.alerts or []:
            other_resources = DomainRelationApply.from_edge_to_resources(
                cache, start_node=self, end_node=alert, edge_type=edge_type, view_by_read_class=view_by_read_class
            )
            resources.extend(other_resources)

        edge_type = dm.DirectRelationReference("power-ops-types", "partialBid")
        for partial in self.partials or []:
            other_resources = DomainRelationApply.from_edge_to_resources(
                cache, start_node=self, end_node=partial, edge_type=edge_type, view_by_read_class=view_by_read_class
            )
            resources.extend(other_resources)

        if isinstance(self.price_area, DomainModelApply):
            other_resources = self.price_area._to_instances_apply(cache, view_by_read_class)
            resources.extend(other_resources)

        if isinstance(self.method, DomainModelApply):
            other_resources = self.method._to_instances_apply(cache, view_by_read_class)
            resources.extend(other_resources)

        if isinstance(self.total, DomainModelApply):
            other_resources = self.total._to_instances_apply(cache, view_by_read_class)
            resources.extend(other_resources)

        return resources


class BidDocumentList(DomainModelList[BidDocument]):
    """List of bid documents in the read version."""

    _INSTANCE = BidDocument

    def as_apply(self) -> BidDocumentApplyList:
        """Convert these read versions of bid document to the writing versions."""
        return BidDocumentApplyList([node.as_apply() for node in self.data])


class BidDocumentApplyList(DomainModelApplyList[BidDocumentApply]):
    """List of bid documents in the writing version."""

    _INSTANCE = BidDocumentApply


def _create_bid_document_filter(
    view_id: dm.ViewId,
    name: str | list[str] | None = None,
    name_prefix: str | None = None,
    min_delivery_date: datetime.date | None = None,
    max_delivery_date: datetime.date | None = None,
    min_start_calculation: datetime.datetime | None = None,
    max_start_calculation: datetime.datetime | None = None,
    min_end_calculation: datetime.datetime | None = None,
    max_end_calculation: datetime.datetime | None = None,
    is_complete: bool | None = None,
    price_area: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
    method: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
    total: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
    external_id_prefix: str | None = None,
    space: str | list[str] | None = None,
    filter: dm.Filter | None = None,
) -> dm.Filter | None:
    filters = []
    if name is not None and isinstance(name, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("name"), value=name))
    if name and isinstance(name, list):
        filters.append(dm.filters.In(view_id.as_property_ref("name"), values=name))
    if name_prefix:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("name"), value=name_prefix))
    if min_delivery_date or max_delivery_date:
        filters.append(
            dm.filters.Range(
                view_id.as_property_ref("deliveryDate"),
                gte=min_delivery_date.isoformat() if min_delivery_date else None,
                lte=max_delivery_date.isoformat() if max_delivery_date else None,
            )
        )
    if min_start_calculation or max_start_calculation:
        filters.append(
            dm.filters.Range(
                view_id.as_property_ref("startCalculation"),
                gte=min_start_calculation.isoformat(timespec="milliseconds") if min_start_calculation else None,
                lte=max_start_calculation.isoformat(timespec="milliseconds") if max_start_calculation else None,
            )
        )
    if min_end_calculation or max_end_calculation:
        filters.append(
            dm.filters.Range(
                view_id.as_property_ref("endCalculation"),
                gte=min_end_calculation.isoformat(timespec="milliseconds") if min_end_calculation else None,
                lte=max_end_calculation.isoformat(timespec="milliseconds") if max_end_calculation else None,
            )
        )
    if is_complete is not None and isinstance(is_complete, bool):
        filters.append(dm.filters.Equals(view_id.as_property_ref("isComplete"), value=is_complete))
    if price_area and isinstance(price_area, str):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("priceArea"), value={"space": DEFAULT_INSTANCE_SPACE, "externalId": price_area}
            )
        )
    if price_area and isinstance(price_area, tuple):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("priceArea"), value={"space": price_area[0], "externalId": price_area[1]}
            )
        )
    if price_area and isinstance(price_area, list) and isinstance(price_area[0], str):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("priceArea"),
                values=[{"space": DEFAULT_INSTANCE_SPACE, "externalId": item} for item in price_area],
            )
        )
    if price_area and isinstance(price_area, list) and isinstance(price_area[0], tuple):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("priceArea"),
                values=[{"space": item[0], "externalId": item[1]} for item in price_area],
            )
        )
    if method and isinstance(method, str):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("method"), value={"space": DEFAULT_INSTANCE_SPACE, "externalId": method}
            )
        )
    if method and isinstance(method, tuple):
        filters.append(
            dm.filters.Equals(view_id.as_property_ref("method"), value={"space": method[0], "externalId": method[1]})
        )
    if method and isinstance(method, list) and isinstance(method[0], str):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("method"),
                values=[{"space": DEFAULT_INSTANCE_SPACE, "externalId": item} for item in method],
            )
        )
    if method and isinstance(method, list) and isinstance(method[0], tuple):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("method"), values=[{"space": item[0], "externalId": item[1]} for item in method]
            )
        )
    if total and isinstance(total, str):
        filters.append(
            dm.filters.Equals(
                view_id.as_property_ref("total"), value={"space": DEFAULT_INSTANCE_SPACE, "externalId": total}
            )
        )
    if total and isinstance(total, tuple):
        filters.append(
            dm.filters.Equals(view_id.as_property_ref("total"), value={"space": total[0], "externalId": total[1]})
        )
    if total and isinstance(total, list) and isinstance(total[0], str):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("total"),
                values=[{"space": DEFAULT_INSTANCE_SPACE, "externalId": item} for item in total],
            )
        )
    if total and isinstance(total, list) and isinstance(total[0], tuple):
        filters.append(
            dm.filters.In(
                view_id.as_property_ref("total"), values=[{"space": item[0], "externalId": item[1]} for item in total]
            )
        )
    if external_id_prefix:
        filters.append(dm.filters.Prefix(["node", "externalId"], value=external_id_prefix))
    if space is not None and isinstance(space, str):
        filters.append(dm.filters.Equals(["node", "space"], value=space))
    if space and isinstance(space, list):
        filters.append(dm.filters.In(["node", "space"], values=space))
    if filter:
        filters.append(filter)
    return dm.filters.And(*filters) if filters else None



from __future__ import annotations

import datetime
from collections.abc import Sequence
from typing import overload

from cognite.client import CogniteClient
from cognite.client import data_modeling as dm
from cognite.client.data_classes.data_modeling.instances import InstanceAggregationResultList

from cognite.powerops.client._generated.day_ahead_bid.data_classes._core import DEFAULT_INSTANCE_SPACE
from cognite.powerops.client._generated.day_ahead_bid.data_classes import (
    DomainModelCore,
    DomainModelApply,
    ResourcesApplyResult,
    BidDocument,
    BidDocumentApply,
    BidDocumentFields,
    BidDocumentList,
    BidDocumentApplyList,
    BidDocumentTextFields,
)
from cognite.powerops.client._generated.day_ahead_bid.data_classes._bid_document import (
    _BIDDOCUMENT_PROPERTIES_BY_FIELD,
    _create_bid_document_filter,
)
from ._core import (
    DEFAULT_LIMIT_READ,
    DEFAULT_QUERY_LIMIT,
    Aggregations,
    NodeAPI,
    SequenceNotStr,
    QueryStep,
    QueryBuilder,
)
from .bid_document_alerts import BidDocumentAlertsAPI
from .bid_document_partials import BidDocumentPartialsAPI
from .bid_document_query import BidDocumentQueryAPI


class BidDocumentAPI(NodeAPI[BidDocument, BidDocumentApply, BidDocumentList]):
    def __init__(self, client: CogniteClient, view_by_read_class: dict[type[DomainModelCore], dm.ViewId]):
        view_id = view_by_read_class[BidDocument]
        super().__init__(
            client=client,
            sources=view_id,
            class_type=BidDocument,
            class_list=BidDocumentList,
            class_apply_list=BidDocumentApplyList,
            view_by_read_class=view_by_read_class,
        )
        self._view_id = view_id
        self.alerts_edge = BidDocumentAlertsAPI(client)
        self.partials_edge = BidDocumentPartialsAPI(client)

    def __call__(
        self,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        min_delivery_date: datetime.date | None = None,
        max_delivery_date: datetime.date | None = None,
        min_start_calculation: datetime.datetime | None = None,
        max_start_calculation: datetime.datetime | None = None,
        min_end_calculation: datetime.datetime | None = None,
        max_end_calculation: datetime.datetime | None = None,
        is_complete: bool | None = None,
        price_area: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        method: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        total: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int = DEFAULT_QUERY_LIMIT,
        filter: dm.Filter | None = None,
    ) -> BidDocumentQueryAPI[BidDocumentList]:
        """Query starting at bid documents.

        Args:
            name: The name to filter on.
            name_prefix: The prefix of the name to filter on.
            min_delivery_date: The minimum value of the delivery date to filter on.
            max_delivery_date: The maximum value of the delivery date to filter on.
            min_start_calculation: The minimum value of the start calculation to filter on.
            max_start_calculation: The maximum value of the start calculation to filter on.
            min_end_calculation: The minimum value of the end calculation to filter on.
            max_end_calculation: The maximum value of the end calculation to filter on.
            is_complete: The is complete to filter on.
            price_area: The price area to filter on.
            method: The method to filter on.
            total: The total to filter on.
            external_id_prefix: The prefix of the external ID to filter on.
            space: The space to filter on.
            limit: Maximum number of bid documents to return. Defaults to 25. Set to -1, float("inf") or None to return all items.
            filter: (Advanced) If the filtering available in the above is not sufficient, you can write your own filtering which will be ANDed with the filter above.

        Returns:
            A query API for bid documents.

        """
        has_data = dm.filters.HasData(views=[self._view_id])
        filter_ = _create_bid_document_filter(
            self._view_id,
            name,
            name_prefix,
            min_delivery_date,
            max_delivery_date,
            min_start_calculation,
            max_start_calculation,
            min_end_calculation,
            max_end_calculation,
            is_complete,
            price_area,
            method,
            total,
            external_id_prefix,
            space,
            (filter and dm.filters.And(filter, has_data)) or has_data,
        )
        builder = QueryBuilder(BidDocumentList)
        return BidDocumentQueryAPI(self._client, builder, self._view_by_read_class, filter_, limit)

    def apply(
        self, bid_document: BidDocumentApply | Sequence[BidDocumentApply], replace: bool = False
    ) -> ResourcesApplyResult:
        """Add or update (upsert) bid documents.

        Note: This method iterates through all nodes and timeseries linked to bid_document and creates them including the edges
        between the nodes. For example, if any of `alerts` or `partials` are set, then these
        nodes as well as any nodes linked to them, and all the edges linking these nodes will be created.

        Args:
            bid_document: Bid document or sequence of bid documents to upsert.
            replace (bool): How do we behave when a property value exists? Do we replace all matching and existing values with the supplied values (true)?
                Or should we merge in new values for properties together with the existing values (false)? Note: This setting applies for all nodes or edges specified in the ingestion call.
        Returns:
            Created instance(s), i.e., nodes, edges, and time series.

        Examples:

            Create a new bid_document:

                >>> from cognite.powerops.client._generated.day_ahead_bid import DayAheadBidAPI
                >>> from cognite.powerops.client._generated.day_ahead_bid.data_classes import BidDocumentApply
                >>> client = DayAheadBidAPI()
                >>> bid_document = BidDocumentApply(external_id="my_bid_document", ...)
                >>> result = client.bid_document.apply(bid_document)

        """
        return self._apply(bid_document, replace)

    def delete(
        self, external_id: str | SequenceNotStr[str], space: str = DEFAULT_INSTANCE_SPACE
    ) -> dm.InstancesDeleteResult:
        """Delete one or more bid document.

        Args:
            external_id: External id of the bid document to delete.
            space: The space where all the bid document are located.

        Returns:
            The instance(s), i.e., nodes and edges which has been deleted. Empty list if nothing was deleted.

        Examples:

            Delete bid_document by id:

                >>> from cognite.powerops.client._generated.day_ahead_bid import DayAheadBidAPI
                >>> client = DayAheadBidAPI()
                >>> client.bid_document.delete("my_bid_document")
        """
        return self._delete(external_id, space)

    @overload
    def retrieve(self, external_id: str, space: str = DEFAULT_INSTANCE_SPACE) -> BidDocument | None:
        ...

    @overload
    def retrieve(self, external_id: SequenceNotStr[str], space: str = DEFAULT_INSTANCE_SPACE) -> BidDocumentList:
        ...

    def retrieve(
        self, external_id: str | SequenceNotStr[str], space: str = DEFAULT_INSTANCE_SPACE
    ) -> BidDocument | BidDocumentList | None:
        """Retrieve one or more bid documents by id(s).

        Args:
            external_id: External id or list of external ids of the bid documents.
            space: The space where all the bid documents are located.

        Returns:
            The requested bid documents.

        Examples:

            Retrieve bid_document by id:

                >>> from cognite.powerops.client._generated.day_ahead_bid import DayAheadBidAPI
                >>> client = DayAheadBidAPI()
                >>> bid_document = client.bid_document.retrieve("my_bid_document")

        """
        return self._retrieve(
            external_id,
            space,
            retrieve_edges=True,
            edge_api_name_type_direction_quad=[
                (
                    self.alerts_edge,
                    "alerts",
                    dm.DirectRelationReference("power-ops-types", "calculationIssue"),
                    "outwards",
                ),
                (
                    self.partials_edge,
                    "partials",
                    dm.DirectRelationReference("power-ops-types", "partialBid"),
                    "outwards",
                ),
            ],
        )

    def search(
        self,
        query: str,
        properties: BidDocumentTextFields | Sequence[BidDocumentTextFields] | None = None,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        min_delivery_date: datetime.date | None = None,
        max_delivery_date: datetime.date | None = None,
        min_start_calculation: datetime.datetime | None = None,
        max_start_calculation: datetime.datetime | None = None,
        min_end_calculation: datetime.datetime | None = None,
        max_end_calculation: datetime.datetime | None = None,
        is_complete: bool | None = None,
        price_area: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        method: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        total: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> BidDocumentList:
        """Search bid documents

        Args:
            query: The search query,
            properties: The property to search, if nothing is passed all text fields will be searched.
            name: The name to filter on.
            name_prefix: The prefix of the name to filter on.
            min_delivery_date: The minimum value of the delivery date to filter on.
            max_delivery_date: The maximum value of the delivery date to filter on.
            min_start_calculation: The minimum value of the start calculation to filter on.
            max_start_calculation: The maximum value of the start calculation to filter on.
            min_end_calculation: The minimum value of the end calculation to filter on.
            max_end_calculation: The maximum value of the end calculation to filter on.
            is_complete: The is complete to filter on.
            price_area: The price area to filter on.
            method: The method to filter on.
            total: The total to filter on.
            external_id_prefix: The prefix of the external ID to filter on.
            space: The space to filter on.
            limit: Maximum number of bid documents to return. Defaults to 25. Set to -1, float("inf") or None to return all items.
            filter: (Advanced) If the filtering available in the above is not sufficient, you can write your own filtering which will be ANDed with the filter above.

        Returns:
            Search results bid documents matching the query.

        Examples:

           Search for 'my_bid_document' in all text properties:

                >>> from cognite.powerops.client._generated.day_ahead_bid import DayAheadBidAPI
                >>> client = DayAheadBidAPI()
                >>> bid_documents = client.bid_document.search('my_bid_document')

        """
        filter_ = _create_bid_document_filter(
            self._view_id,
            name,
            name_prefix,
            min_delivery_date,
            max_delivery_date,
            min_start_calculation,
            max_start_calculation,
            min_end_calculation,
            max_end_calculation,
            is_complete,
            price_area,
            method,
            total,
            external_id_prefix,
            space,
            filter,
        )
        return self._search(self._view_id, query, _BIDDOCUMENT_PROPERTIES_BY_FIELD, properties, filter_, limit)

    @overload
    def aggregate(
        self,
        aggregations: Aggregations
        | dm.aggregations.MetricAggregation
        | Sequence[Aggregations]
        | Sequence[dm.aggregations.MetricAggregation],
        property: BidDocumentFields | Sequence[BidDocumentFields] | None = None,
        group_by: None = None,
        query: str | None = None,
        search_properties: BidDocumentTextFields | Sequence[BidDocumentTextFields] | None = None,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        min_delivery_date: datetime.date | None = None,
        max_delivery_date: datetime.date | None = None,
        min_start_calculation: datetime.datetime | None = None,
        max_start_calculation: datetime.datetime | None = None,
        min_end_calculation: datetime.datetime | None = None,
        max_end_calculation: datetime.datetime | None = None,
        is_complete: bool | None = None,
        price_area: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        method: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        total: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> list[dm.aggregations.AggregatedNumberedValue]:
        ...

    @overload
    def aggregate(
        self,
        aggregations: Aggregations
        | dm.aggregations.MetricAggregation
        | Sequence[Aggregations]
        | Sequence[dm.aggregations.MetricAggregation],
        property: BidDocumentFields | Sequence[BidDocumentFields] | None = None,
        group_by: BidDocumentFields | Sequence[BidDocumentFields] = None,
        query: str | None = None,
        search_properties: BidDocumentTextFields | Sequence[BidDocumentTextFields] | None = None,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        min_delivery_date: datetime.date | None = None,
        max_delivery_date: datetime.date | None = None,
        min_start_calculation: datetime.datetime | None = None,
        max_start_calculation: datetime.datetime | None = None,
        min_end_calculation: datetime.datetime | None = None,
        max_end_calculation: datetime.datetime | None = None,
        is_complete: bool | None = None,
        price_area: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        method: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        total: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> InstanceAggregationResultList:
        ...

    def aggregate(
        self,
        aggregate: Aggregations
        | dm.aggregations.MetricAggregation
        | Sequence[Aggregations]
        | Sequence[dm.aggregations.MetricAggregation],
        property: BidDocumentFields | Sequence[BidDocumentFields] | None = None,
        group_by: BidDocumentFields | Sequence[BidDocumentFields] | None = None,
        query: str | None = None,
        search_property: BidDocumentTextFields | Sequence[BidDocumentTextFields] | None = None,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        min_delivery_date: datetime.date | None = None,
        max_delivery_date: datetime.date | None = None,
        min_start_calculation: datetime.datetime | None = None,
        max_start_calculation: datetime.datetime | None = None,
        min_end_calculation: datetime.datetime | None = None,
        max_end_calculation: datetime.datetime | None = None,
        is_complete: bool | None = None,
        price_area: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        method: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        total: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> list[dm.aggregations.AggregatedNumberedValue] | InstanceAggregationResultList:
        """Aggregate data across bid documents

        Args:
            aggregate: The aggregation to perform.
            property: The property to perform aggregation on.
            group_by: The property to group by when doing the aggregation.
            query: The query to search for in the text field.
            search_property: The text field to search in.
            name: The name to filter on.
            name_prefix: The prefix of the name to filter on.
            min_delivery_date: The minimum value of the delivery date to filter on.
            max_delivery_date: The maximum value of the delivery date to filter on.
            min_start_calculation: The minimum value of the start calculation to filter on.
            max_start_calculation: The maximum value of the start calculation to filter on.
            min_end_calculation: The minimum value of the end calculation to filter on.
            max_end_calculation: The maximum value of the end calculation to filter on.
            is_complete: The is complete to filter on.
            price_area: The price area to filter on.
            method: The method to filter on.
            total: The total to filter on.
            external_id_prefix: The prefix of the external ID to filter on.
            space: The space to filter on.
            limit: Maximum number of bid documents to return. Defaults to 25. Set to -1, float("inf") or None to return all items.
            filter: (Advanced) If the filtering available in the above is not sufficient, you can write your own filtering which will be ANDed with the filter above.

        Returns:
            Aggregation results.

        Examples:

            Count bid documents in space `my_space`:

                >>> from cognite.powerops.client._generated.day_ahead_bid import DayAheadBidAPI
                >>> client = DayAheadBidAPI()
                >>> result = client.bid_document.aggregate("count", space="my_space")

        """

        filter_ = _create_bid_document_filter(
            self._view_id,
            name,
            name_prefix,
            min_delivery_date,
            max_delivery_date,
            min_start_calculation,
            max_start_calculation,
            min_end_calculation,
            max_end_calculation,
            is_complete,
            price_area,
            method,
            total,
            external_id_prefix,
            space,
            filter,
        )
        return self._aggregate(
            self._view_id,
            aggregate,
            _BIDDOCUMENT_PROPERTIES_BY_FIELD,
            property,
            group_by,
            query,
            search_property,
            limit,
            filter_,
        )

    def histogram(
        self,
        property: BidDocumentFields,
        interval: float,
        query: str | None = None,
        search_property: BidDocumentTextFields | Sequence[BidDocumentTextFields] | None = None,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        min_delivery_date: datetime.date | None = None,
        max_delivery_date: datetime.date | None = None,
        min_start_calculation: datetime.datetime | None = None,
        max_start_calculation: datetime.datetime | None = None,
        min_end_calculation: datetime.datetime | None = None,
        max_end_calculation: datetime.datetime | None = None,
        is_complete: bool | None = None,
        price_area: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        method: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        total: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
    ) -> dm.aggregations.HistogramValue:
        """Produces histograms for bid documents

        Args:
            property: The property to use as the value in the histogram.
            interval: The interval to use for the histogram bins.
            query: The query to search for in the text field.
            search_property: The text field to search in.
            name: The name to filter on.
            name_prefix: The prefix of the name to filter on.
            min_delivery_date: The minimum value of the delivery date to filter on.
            max_delivery_date: The maximum value of the delivery date to filter on.
            min_start_calculation: The minimum value of the start calculation to filter on.
            max_start_calculation: The maximum value of the start calculation to filter on.
            min_end_calculation: The minimum value of the end calculation to filter on.
            max_end_calculation: The maximum value of the end calculation to filter on.
            is_complete: The is complete to filter on.
            price_area: The price area to filter on.
            method: The method to filter on.
            total: The total to filter on.
            external_id_prefix: The prefix of the external ID to filter on.
            space: The space to filter on.
            limit: Maximum number of bid documents to return. Defaults to 25. Set to -1, float("inf") or None to return all items.
            filter: (Advanced) If the filtering available in the above is not sufficient, you can write your own filtering which will be ANDed with the filter above.

        Returns:
            Bucketed histogram results.

        """
        filter_ = _create_bid_document_filter(
            self._view_id,
            name,
            name_prefix,
            min_delivery_date,
            max_delivery_date,
            min_start_calculation,
            max_start_calculation,
            min_end_calculation,
            max_end_calculation,
            is_complete,
            price_area,
            method,
            total,
            external_id_prefix,
            space,
            filter,
        )
        return self._histogram(
            self._view_id,
            property,
            interval,
            _BIDDOCUMENT_PROPERTIES_BY_FIELD,
            query,
            search_property,
            limit,
            filter_,
        )

    def list(
        self,
        name: str | list[str] | None = None,
        name_prefix: str | None = None,
        min_delivery_date: datetime.date | None = None,
        max_delivery_date: datetime.date | None = None,
        min_start_calculation: datetime.datetime | None = None,
        max_start_calculation: datetime.datetime | None = None,
        min_end_calculation: datetime.datetime | None = None,
        max_end_calculation: datetime.datetime | None = None,
        is_complete: bool | None = None,
        price_area: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        method: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        total: str | tuple[str, str] | list[str] | list[tuple[str, str]] | None = None,
        external_id_prefix: str | None = None,
        space: str | list[str] | None = None,
        limit: int = DEFAULT_LIMIT_READ,
        filter: dm.Filter | None = None,
        retrieve_edges: bool = True,
    ) -> BidDocumentList:
        """List/filter bid documents

        Args:
            name: The name to filter on.
            name_prefix: The prefix of the name to filter on.
            min_delivery_date: The minimum value of the delivery date to filter on.
            max_delivery_date: The maximum value of the delivery date to filter on.
            min_start_calculation: The minimum value of the start calculation to filter on.
            max_start_calculation: The maximum value of the start calculation to filter on.
            min_end_calculation: The minimum value of the end calculation to filter on.
            max_end_calculation: The maximum value of the end calculation to filter on.
            is_complete: The is complete to filter on.
            price_area: The price area to filter on.
            method: The method to filter on.
            total: The total to filter on.
            external_id_prefix: The prefix of the external ID to filter on.
            space: The space to filter on.
            limit: Maximum number of bid documents to return. Defaults to 25. Set to -1, float("inf") or None to return all items.
            filter: (Advanced) If the filtering available in the above is not sufficient, you can write your own filtering which will be ANDed with the filter above.
            retrieve_edges: Whether to retrieve `alerts` or `partials` external ids for the bid documents. Defaults to True.

        Returns:
            List of requested bid documents

        Examples:

            List bid documents and limit to 5:

                >>> from cognite.powerops.client._generated.day_ahead_bid import DayAheadBidAPI
                >>> client = DayAheadBidAPI()
                >>> bid_documents = client.bid_document.list(limit=5)

        """
        filter_ = _create_bid_document_filter(
            self._view_id,
            name,
            name_prefix,
            min_delivery_date,
            max_delivery_date,
            min_start_calculation,
            max_start_calculation,
            min_end_calculation,
            max_end_calculation,
            is_complete,
            price_area,
            method,
            total,
            external_id_prefix,
            space,
            filter,
        )

        return self._list(
            limit=limit,
            filter=filter_,
            retrieve_edges=retrieve_edges,
            edge_api_name_type_direction_quad=[
                (
                    self.alerts_edge,
                    "alerts",
                    dm.DirectRelationReference("power-ops-types", "calculationIssue"),
                    "outwards",
                ),
                (
                    self.partials_edge,
                    "partials",
                    dm.DirectRelationReference("power-ops-types", "partialBid"),
                    "outwards",
                ),
            ],
        )



from ._core import (
    DomainModel,
    DomainModelCore,
    DomainModelApply,
    DomainModelList,
    DomainRelationApply,
    ResourcesApply,
    ResourcesApplyResult,
)
from ._alert import Alert, AlertApply, AlertApplyList, AlertFields, AlertList, AlertTextFields
from ._bid_document import (
    BidDocument,
    BidDocumentApply,
    BidDocumentApplyList,
    BidDocumentFields,
    BidDocumentList,
    BidDocumentTextFields,
)
from ._bid_method import (
    BidMethod,
    BidMethodApply,
    BidMethodApplyList,
    BidMethodFields,
    BidMethodList,
    BidMethodTextFields,
)
from ._bid_row import BidRow, BidRowApply, BidRowApplyList, BidRowFields, BidRowList, BidRowTextFields
from ._price_area import PriceArea, PriceAreaFields, PriceAreaList, PriceAreaTextFields

BidDocument.model_rebuild()
BidDocumentApply.model_rebuild()
BidRow.model_rebuild()
BidRowApply.model_rebuild()

__all__ = [
    "ResourcesApply",
    "DomainModel",
    "DomainModelCore",
    "DomainModelApply",
    "DomainModelList",
    "DomainRelationApply",
    "ResourcesApplyResult",
    "Alert",
    "AlertApply",
    "AlertList",
    "AlertApplyList",
    "AlertFields",
    "AlertTextFields",
    "BidDocument",
    "BidDocumentApply",
    "BidDocumentList",
    "BidDocumentApplyList",
    "BidDocumentFields",
    "BidDocumentTextFields",
    "BidMethod",
    "BidMethodApply",
    "BidMethodList",
    "BidMethodApplyList",
    "BidMethodFields",
    "BidMethodTextFields",
    "BidRow",
    "BidRowApply",
    "BidRowList",
    "BidRowApplyList",
    "BidRowFields",
    "BidRowTextFields",
    "PriceArea",
    "PriceAreaList",
    "PriceAreaFields",
    "PriceAreaTextFields",
]



from __future__ import annotations

from typing import Literal, Optional, Union

from cognite.client import data_modeling as dm

from ._core import (
    DEFAULT_INSTANCE_SPACE,
    DomainModel,
    DomainModelCore,
    DomainModelApply,
    DomainModelApplyList,
    DomainModelList,
    DomainRelationApply,
    ResourcesApply,
)


__all__ = [
    "GeneratorEfficiencyCurve",
    "GeneratorEfficiencyCurveApply",
    "GeneratorEfficiencyCurveList",
    "GeneratorEfficiencyCurveApplyList",
    "GeneratorEfficiencyCurveFields",
]

GeneratorEfficiencyCurveFields = Literal["ref", "power", "efficiency"]

_GENERATOREFFICIENCYCURVE_PROPERTIES_BY_FIELD = {
    "ref": "ref",
    "power": "power",
    "efficiency": "efficiency",
}


class GeneratorEfficiencyCurve(DomainModel):
    """This represents the reading version of generator efficiency curve.

    It is used to when data is retrieved from CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the generator efficiency curve.
        ref: The reference value
        power: The generator power values
        efficiency: The generator efficiency values
        created_time: The created time of the generator efficiency curve node.
        last_updated_time: The last updated time of the generator efficiency curve node.
        deleted_time: If present, the deleted time of the generator efficiency curve node.
        version: The version of the generator efficiency curve node.
    """

    space: str = DEFAULT_INSTANCE_SPACE
    node_type: Union[dm.DirectRelationReference, None] = dm.DirectRelationReference(
        "power-ops-types", "GeneratorEfficiencyCurve"
    )
    ref: Optional[float] = None
    power: Optional[list[float]] = None
    efficiency: Optional[list[float]] = None

    def as_apply(self) -> GeneratorEfficiencyCurveApply:
        """Convert this read version of generator efficiency curve to the writing version."""
        return GeneratorEfficiencyCurveApply(
            space=self.space,
            external_id=self.external_id,
            ref=self.ref,
            power=self.power,
            efficiency=self.efficiency,
        )


class GeneratorEfficiencyCurveApply(DomainModelApply):
    """This represents the writing version of generator efficiency curve.

    It is used to when data is sent to CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the generator efficiency curve.
        ref: The reference value
        power: The generator power values
        efficiency: The generator efficiency values
        existing_version: Fail the ingestion request if the generator efficiency curve version is greater than or equal to this value.
            If no existingVersion is specified, the ingestion will always overwrite any existing data for the edge (for the specified container or instance).
            If existingVersion is set to 0, the upsert will behave as an insert, so it will fail the bulk if the item already exists.
            If skipOnVersionConflict is set on the ingestion request, then the item will be skipped instead of failing the ingestion request.
    """

    space: str = DEFAULT_INSTANCE_SPACE
    node_type: Union[dm.DirectRelationReference, None] = dm.DirectRelationReference(
        "power-ops-types", "GeneratorEfficiencyCurve"
    )
    ref: Optional[float] = None
    power: list[float]
    efficiency: list[float]

    def _to_instances_apply(
        self,
        cache: set[tuple[str, str]],
        view_by_read_class: dict[type[DomainModelCore], dm.ViewId] | None,
    ) -> ResourcesApply:
        resources = ResourcesApply()
        if self.as_tuple_id() in cache:
            return resources

        write_view = (view_by_read_class or {}).get(
            GeneratorEfficiencyCurve, dm.ViewId("power-ops-assets", "GeneratorEfficiencyCurve", "1")
        )

        properties = {}

        if self.ref is not None:
            properties["ref"] = self.ref

        if self.power is not None:
            properties["power"] = self.power

        if self.efficiency is not None:
            properties["efficiency"] = self.efficiency

        if properties:
            this_node = dm.NodeApply(
                space=self.space,
                external_id=self.external_id,
                existing_version=self.existing_version,
                type=self.node_type,
                sources=[
                    dm.NodeOrEdgeData(
                        source=write_view,
                        properties=properties,
                    )
                ],
            )
            resources.nodes.append(this_node)
            cache.add(self.as_tuple_id())

        return resources


class GeneratorEfficiencyCurveList(DomainModelList[GeneratorEfficiencyCurve]):
    """List of generator efficiency curves in the read version."""

    _INSTANCE = GeneratorEfficiencyCurve

    def as_apply(self) -> GeneratorEfficiencyCurveApplyList:
        """Convert these read versions of generator efficiency curve to the writing versions."""
        return GeneratorEfficiencyCurveApplyList([node.as_apply() for node in self.data])


class GeneratorEfficiencyCurveApplyList(DomainModelApplyList[GeneratorEfficiencyCurveApply]):
    """List of generator efficiency curves in the writing version."""

    _INSTANCE = GeneratorEfficiencyCurveApply


def _create_generator_efficiency_curve_filter(
    view_id: dm.ViewId,
    min_ref: float | None = None,
    max_ref: float | None = None,
    external_id_prefix: str | None = None,
    space: str | list[str] | None = None,
    filter: dm.Filter | None = None,
) -> dm.Filter | None:
    filters = []
    if min_ref or max_ref:
        filters.append(dm.filters.Range(view_id.as_property_ref("ref"), gte=min_ref, lte=max_ref))
    if external_id_prefix:
        filters.append(dm.filters.Prefix(["node", "externalId"], value=external_id_prefix))
    if space is not None and isinstance(space, str):
        filters.append(dm.filters.Equals(["node", "space"], value=space))
    if space and isinstance(space, list):
        filters.append(dm.filters.In(["node", "space"], values=space))
    if filter:
        filters.append(filter)
    return dm.filters.And(*filters) if filters else None



from __future__ import annotations

from typing import Literal, Optional, Union

from cognite.client import data_modeling as dm
from pydantic import Field

from ._core import (
    DEFAULT_INSTANCE_SPACE,
    DomainModel,
    DomainModelCore,
    DomainModelApply,
    DomainModelApplyList,
    DomainModelList,
    DomainRelationApply,
    ResourcesApply,
)


__all__ = [
    "Reservoir",
    "ReservoirApply",
    "ReservoirList",
    "ReservoirApplyList",
    "ReservoirFields",
    "ReservoirTextFields",
]


ReservoirTextFields = Literal["name", "display_name"]
ReservoirFields = Literal["name", "display_name", "ordering"]

_RESERVOIR_PROPERTIES_BY_FIELD = {
    "name": "name",
    "display_name": "displayName",
    "ordering": "ordering",
}


class Reservoir(DomainModel):
    """This represents the reading version of reservoir.

    It is used to when data is retrieved from CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the reservoir.
        name: The name field.
        display_name: The display name field.
        ordering: The ordering field.
        created_time: The created time of the reservoir node.
        last_updated_time: The last updated time of the reservoir node.
        deleted_time: If present, the deleted time of the reservoir node.
        version: The version of the reservoir node.
    """

    space: str = DEFAULT_INSTANCE_SPACE
    node_type: Union[dm.DirectRelationReference, None] = None
    name: Optional[str] = None
    display_name: Optional[str] = Field(None, alias="displayName")
    ordering: Optional[int] = None

    def as_apply(self) -> ReservoirApply:
        """Convert this read version of reservoir to the writing version."""
        return ReservoirApply(
            space=self.space,
            external_id=self.external_id,
            name=self.name,
            display_name=self.display_name,
            ordering=self.ordering,
        )


class ReservoirApply(DomainModelApply):
    """This represents the writing version of reservoir.

    It is used to when data is sent to CDF.

    Args:
        space: The space where the node is located.
        external_id: The external id of the reservoir.
        name: The name field.
        display_name: The display name field.
        ordering: The ordering field.
        existing_version: Fail the ingestion request if the reservoir version is greater than or equal to this value.
            If no existingVersion is specified, the ingestion will always overwrite any existing data for the edge (for the specified container or instance).
            If existingVersion is set to 0, the upsert will behave as an insert, so it will fail the bulk if the item already exists.
            If skipOnVersionConflict is set on the ingestion request, then the item will be skipped instead of failing the ingestion request.
    """

    space: str = DEFAULT_INSTANCE_SPACE
    node_type: Union[dm.DirectRelationReference, None] = None
    name: Optional[str] = None
    display_name: Optional[str] = Field(None, alias="displayName")
    ordering: Optional[int] = None

    def _to_instances_apply(
        self,
        cache: set[tuple[str, str]],
        view_by_read_class: dict[type[DomainModelCore], dm.ViewId] | None,
    ) -> ResourcesApply:
        resources = ResourcesApply()
        if self.as_tuple_id() in cache:
            return resources

        write_view = (view_by_read_class or {}).get(Reservoir, dm.ViewId("power-ops", "Reservoir", "3c822b0c3d68f7"))

        properties = {}

        if self.name is not None:
            properties["name"] = self.name

        if self.display_name is not None:
            properties["displayName"] = self.display_name

        if self.ordering is not None:
            properties["ordering"] = self.ordering

        if properties:
            this_node = dm.NodeApply(
                space=self.space,
                external_id=self.external_id,
                existing_version=self.existing_version,
                type=self.node_type,
                sources=[
                    dm.NodeOrEdgeData(
                        source=write_view,
                        properties=properties,
                    )
                ],
            )
            resources.nodes.append(this_node)
            cache.add(self.as_tuple_id())

        return resources


class ReservoirList(DomainModelList[Reservoir]):
    """List of reservoirs in the read version."""

    _INSTANCE = Reservoir

    def as_apply(self) -> ReservoirApplyList:
        """Convert these read versions of reservoir to the writing versions."""
        return ReservoirApplyList([node.as_apply() for node in self.data])


class ReservoirApplyList(DomainModelApplyList[ReservoirApply]):
    """List of reservoirs in the writing version."""

    _INSTANCE = ReservoirApply


def _create_reservoir_filter(
    view_id: dm.ViewId,
    name: str | list[str] | None = None,
    name_prefix: str | None = None,
    display_name: str | list[str] | None = None,
    display_name_prefix: str | None = None,
    min_ordering: int | None = None,
    max_ordering: int | None = None,
    external_id_prefix: str | None = None,
    space: str | list[str] | None = None,
    filter: dm.Filter | None = None,
) -> dm.Filter | None:
    filters = []
    if name is not None and isinstance(name, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("name"), value=name))
    if name and isinstance(name, list):
        filters.append(dm.filters.In(view_id.as_property_ref("name"), values=name))
    if name_prefix:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("name"), value=name_prefix))
    if display_name is not None and isinstance(display_name, str):
        filters.append(dm.filters.Equals(view_id.as_property_ref("displayName"), value=display_name))
    if display_name and isinstance(display_name, list):
        filters.append(dm.filters.In(view_id.as_property_ref("displayName"), values=display_name))
    if display_name_prefix:
        filters.append(dm.filters.Prefix(view_id.as_property_ref("displayName"), value=display_name_prefix))
    if min_ordering or max_ordering:
        filters.append(dm.filters.Range(view_id.as_property_ref("ordering"), gte=min_ordering, lte=max_ordering))
    if external_id_prefix:
        filters.append(dm.filters.Prefix(["node", "externalId"], value=external_id_prefix))
    if space is not None and isinstance(space, str):
        filters.append(dm.filters.Equals(["node", "space"], value=space))
    if space and isinstance(space, list):
        filters.append(dm.filters.In(["node", "space"], values=space))
    if filter:
        filters.append(filter)
    return dm.filters.And(*filters) if filters else None



from __future__ import annotations

from pathlib import Path

from cognite.client import ClientConfig, CogniteClient, data_modeling as dm
from cognite.client.credentials import OAuthClientCredentials

from ._api.generator import GeneratorAPI
from ._api.plant import PlantAPI
from ._api.price_area import PriceAreaAPI
from ._api.reservoir import ReservoirAPI
from ._api.watercourse import WatercourseAPI
from ._api.watercourse_shop import WatercourseShopAPI
from . import data_classes


class ProductionModelAPI:
    """
    ProductionModelAPI

    Generated with:
        pygen = 0.33.0
        cognite-sdk = 7.8.6
        pydantic = 2.5.3

    Data Model:
        space: power-ops
        externalId: production
        version: 1
    """

    def __init__(self, config_or_client: CogniteClient | ClientConfig):
        if isinstance(config_or_client, CogniteClient):
            client = config_or_client
        elif isinstance(config_or_client, ClientConfig):
            client = CogniteClient(config_or_client)
        else:
            raise ValueError(f"Expected CogniteClient or ClientConfig, got {type(config_or_client)}")
        # The client name is used for aggregated logging of Pygen Usage
        client.config.client_name = "CognitePygen:0.33.0"

        view_by_read_class = {
            data_classes.Generator: dm.ViewId("power-ops", "Generator", "9178931bbaac71"),
            data_classes.Plant: dm.ViewId("power-ops", "Plant", "836dcb3f5da1df"),
            data_classes.PriceArea: dm.ViewId("power-ops", "PriceArea", "6849ae787cd368"),
            data_classes.Reservoir: dm.ViewId("power-ops", "Reservoir", "3c822b0c3d68f7"),
            data_classes.Watercourse: dm.ViewId("power-ops", "Watercourse", "96f5170f35ef70"),
            data_classes.WatercourseShop: dm.ViewId("power-ops", "WatercourseShop", "4b5321b1fccd06"),
        }

        self.generator = GeneratorAPI(client, view_by_read_class)
        self.plant = PlantAPI(client, view_by_read_class)
        self.price_area = PriceAreaAPI(client, view_by_read_class)
        self.reservoir = ReservoirAPI(client, view_by_read_class)
        self.watercourse = WatercourseAPI(client, view_by_read_class)
        self.watercourse_shop = WatercourseShopAPI(client, view_by_read_class)

    @classmethod
    def azure_project(
        cls, tenant_id: str, client_id: str, client_secret: str, cdf_cluster: str, project: str
    ) -> ProductionModelAPI:
        credentials = OAuthClientCredentials.default_for_azure_ad(tenant_id, client_id, client_secret, cdf_cluster)
        config = ClientConfig.default(project, cdf_cluster, credentials)

        return cls(config)

    @classmethod
    def from_toml(cls, file_path: Path | str, section: str | None = "cognite") -> ProductionModelAPI:
        import toml

        toml_content = toml.load(file_path)
        if section is not None:
            try:
                toml_content = toml_content[section]
            except KeyError as e:
                raise ValueError(f"Could not find section '{section}' in {file_path}") from e

        return cls.azure_project(**toml_content)
